'''
The perturbations module allows to perform the pathway and compounds in-silico perturbations and obatin the perturbation results.
'''

from .compounds import get_top_compounds
from .pathways import get_top_pathways